#!/usr/bin/perl -w

package Databaze;

BEGIN {
    use src::MymoneyUI;
};


use strict;
use DBI;

use Encode;
use utf8;
use open ':utf8', ':std';

my %MyMoney;
my %STH;

sub new {
  my $class = shift;
  my $self = {};
  bless $self, $class;
  %MyMoney=@_;
  $self->{'dbh'}=0;
  $self->{'connected'}=0;
  $self->{'transaction'}=0;
  return $self;
};

sub DBTest {
    my ($self, $MainCfg)=@_;

    if($MainCfg->get('dbhost')) {
        if(!DBI->connect("dbi:".$MainCfg->get('dbtype').":host=".$MainCfg->get('dbhost').";dbname=".$MainCfg->get('dbname'),$MainCfg->get('dbuser'),$MainCfg->get('dbpasswd'))) {
            my $Dialog=new Dialog;
            $Dialog->show("type"=>"error","text"=>"Nezdařilo se připojení k databázi!");
            return(0);
        } else {
            return(1);
        };
    };
    return(0);
};


sub connect {
    my ($self)=@_;
    
    my $MainCfg=$MyMoney{'MainCfg'};
    if(!($self->{'dbh'}=DBI->connect("dbi:".$MainCfg->get('dbtype').":host=".$MainCfg->get('dbhost').";dbname=".$MainCfg->get('dbname'),$MainCfg->get('dbuser'),$MainCfg->get('dbpasswd')))) {
        my $Dialog=new Dialog;
        $Dialog->show("type"=>"error","text"=>"Nezdařilo se připojení k databázi!");
        return(-1);
    };
    $self->{'connected'}=1;
    return(1);
};

# pridame slashe pred znaky, ktere by mohly byt pri praci s DB kolizni
sub ParseStr {
    my ($self,$str)=@_;
    $str=~s/\\/\\\\/g;
    $str=~s/'/\\'/g;
    $str=~s/"/\\"/g;
    return($str);
};

sub begin_work {
    my ($self)=@_;
    $self->{'dbh'}->begin_work;
    $self->{'transaction'}=1;
};

sub commit {
    my ($self)=@_;
    $self->{'dbh'}->commit;
    $self->{'transaction'}=0;
};

sub rollback {
    my ($self)=@_;
    $self->{'dbh'}->rollback;
    $self->{'transaction'}=0;
};

sub lock_tables {
    my ($self,@tables)=@_;
    my $MainCfg=$MyMoney{'MainCfg'};
    my $mysql_locking="";
    foreach $_ (@tables) {
        if($MainCfg->get('dbtype') eq 'Pg') {
            tr/([A-Z])/([a-z])/;
            my $sth=$self->{'dbh'}->prepare("LOCK TABLE ".$_);
            $sth->execute;
        };
        if($MainCfg->get('dbtype') eq 'mysql') {
            if($mysql_locking) {
                $mysql_locking.=", ";
            };
            $mysql_locking.=$_." WRITE";
        };
    };
    if($MainCfg->get('dbtype') eq 'mysql') {
        my $sth=$self->{'dbh'}->prepare("LOCK TABLES ".$mysql_locking);
        $sth->execute;
    };
};

sub unlock_tables {
    my ($self,@tables)=@_;
    my $MainCfg=$MyMoney{'MainCfg'};

    if($MainCfg->get('dbtype') eq 'mysql') {
        my $sth=$self->{'dbh'}->prepare("UNLOCK TABLES");
        $sth->execute;
    };
};

sub Insert {
    my($self,$table,$auto,%data)=@_;

    my $sql_struct="";
    my $sql_data="";
    foreach $_ (keys(%data)) {
        if($sql_struct) {
            $sql_struct.=", ";
        };
        $sql_struct.=$_;
        if($sql_data) {
            $sql_data.=", ";
        };
        #$sql_data.="'".Encode::encode('iso-8859-2', $data{$_})."'";
        $sql_data.="'".$self->ParseStr($data{$_})."'";
    };


    my $MainCfg=$MyMoney{'MainCfg'};
    if($MainCfg->get('dbtype') eq 'Pg') {
        $table=~tr/([A-Z])/([a-z])/;
    };
    if(!$self->{'connected'}) {
        if(!$self->connect) {
            return(-1);
        };
    };
    my $local_transaction=0;
    if(!$self->{'transaction'}) {
        $self->begin_work;
        $self->lock_tables($table);
        $local_transaction=1;
    };
    #print("INSERT INTO ".$table." (".$sql_struct.") VALUES(".$sql_data.");\n");
    $self->{'dbh'}->do("INSERT INTO ".$table." (".$sql_struct.") VALUES(".$sql_data.")");
    if($self->{'dbh'}->err) {
        if($local_transaction) {
            $self->rollback;
            $self->unlock_tables($table);
        };
        return(-1);
    };
    my $id=0;
    if($auto) {
        if($MainCfg->get('dbtype') eq 'mysql') {
            my $sth=$self->{'dbh'}->prepare("SELECT LAST_INSERT_ID()");
            $sth->execute;
            my @row=$sth->fetchrow_array;
            $id=$row[0];
        };
        if($MainCfg->get('dbtype') eq 'Pg') {
            my $sth=$self->{'dbh'}->prepare("SELECT last_value FROM ".$table."_".$auto."_seq");
            $sth->execute;
            my @row=$sth->fetchrow_array;
            $id=$row[0];
        };
    };
    if($local_transaction) {
        $self->commit;
        $self->unlock_tables($table);
    };
    return($id);
};


sub Read {
    my($self, $table,$where,@data)=@_;

    my $sql_struct="";
    foreach $_ (@data) {
        if($sql_struct) {
            $sql_struct.=", ";
        };
        $sql_struct.=$_;
    };
    
    my $MainCfg=$MyMoney{'MainCfg'};
    if($MainCfg->get('dbtype') eq 'Pg') {
        $table=~tr/([A-Z])/([a-z])/;
    };
    if(!$self->{'connected'}) {
        if(!$self->connect) {
            return(-1);
        };
    };
    my $local_transaction=0;
    if(!$self->{'transaction'}) {
        $self->begin_work;
        $self->lock_tables($table);
        $local_transaction=1;
    };

    if($where) {
        #$where=Encode::encode('iso-8859-2', $where);
        #print("SELECT ".$sql_struct." FROM ".$table." WHERE ".$where.";\n");
        $STH{$self}=$self->{'dbh'}->prepare("SELECT ".$sql_struct." FROM ".$table." WHERE ".$where);
    } else {
        $STH{$self}=$self->{'dbh'}->prepare("SELECT ".$sql_struct." FROM ".$table);
    };
    $STH{$self}->execute;
    #$dbh->do("UNLOCK TABLES");
    #print("$table: ".$STH{$self}->rows."\n");
    if($local_transaction) {
        $self->commit;
        $self->unlock_tables($table);
    };
    return($STH{$self}->rows);
};

sub Remove {
    my($self, $table,$where)=@_;

    my $MainCfg=$MyMoney{'MainCfg'};
    if($MainCfg->get('dbtype') eq 'Pg') {
        $table=~tr/([A-Z])/([a-z])/;
    };
    if(!$self->{'connected'}) {
        if(!$self->connect) {
            return(-1);
        };
    };
    my $local_transaction=0;
    if(!$self->{'transaction'}) {
        $self->begin_work;
        $self->lock_tables($table);
        $local_transaction=1;
    };

    #$dbh->do("LOCK TABLES ".$table." WRITE");
    if($where) {
        #$where=Encode::encode('iso-8859-2', $where);
        $STH{$self}=$self->{'dbh'}->prepare("DELETE FROM ".$table." WHERE ".$where);
        $STH{$self}->execute;
    } else {
        print("Error: pokus o smazani vsech dat z tabulky!\n");
        exit;
    };
    #$dbh->do("UNLOCK TABLES");
    if($local_transaction) {
        $self->commit;
        $self->unlock_tables($table);
    };

    return($STH{$self}->rows);
};

# po update je problem s cestinou - prozatim jej zkusime vyresit smazanim radku a novym insertem :(
sub Update {
    my($self, $table,$where,%data)=@_;

    my $sql_data="";
    foreach $_ (keys(%data)) {
        if($sql_data) {
            $sql_data.=", ";
        };
        #$sql_data.=$_."='".Encode::encode('iso-8859-2', $data{$_})."'";
        $sql_data.=$_."='".$self->ParseStr($data{$_})."'";
    };
    
    my $MainCfg=$MyMoney{'MainCfg'};
    if($MainCfg->get('dbtype') eq 'Pg') {
        $table=~tr/([A-Z])/([a-z])/;
    };
    if(!$self->{'connected'}) {
        if(!$self->connect) {
            return(-1);
        };
    };
    my $local_transaction=0;
    if(!$self->{'transaction'}) {
        $self->begin_work;
        $self->lock_tables($table);
        $local_transaction=1;
    };

    my $sth=0;
    if($where) {
        #print("UPDATE ".$table." SET ".$sql_data." WHERE ".$where."\n");
        $sth=$self->{'dbh'}->prepare("UPDATE ".$table." SET ".$sql_data." WHERE ".$where);
    } else {
        #print("UPDATE ".$table." SET ".$sql_data."\n");
        $sth=$self->{'dbh'}->prepare("UPDATE ".$table." SET ".$sql_data);
    };
    my $rc=$sth->execute;
    if($local_transaction) {
        $self->commit;
        $self->unlock_tables($table);
    };
    return($rc);
};

sub numrows {
    my($self)=shift;
    return($STH{$self}->rows);
};

#sub dataseek {
#    my($self)=shift;
#    my($seek)=@_;
#    return($STH{$self}->dataseek($seek));
#};

sub fetchhash {
    my($self)=shift;
    my %row=%{$STH{$self}->fetchrow_hashref};
    foreach $_ (keys(%row)) {
        $row{$_}=Encode::decode('utf8', $row{$_});
    };
    return(%row);
};

sub fetchrow {
    my($self)=shift;

    my @row=$STH{$self}->fetchrow;
    for(my $i=0; $i<@row; $i++ ) {
        $row[$i]=Encode::decode('utf8', $row[$i]);
    };
    return(@row);
};

sub Check {
    my($self, $table)=@_;
    
    my $MainCfg=$MyMoney{'MainCfg'};
    if($MainCfg->get('dbtype') eq 'Pg') {
        $table=~tr/([A-Z])/([a-z])/;
    };
    if(!$self->{'connected'}) {
        if(!$self->connect) {
            return(-1);
        };
    };

    my $sth;
    if($MainCfg->get('dbtype') eq 'mysql') {
        $sth=$self->{'dbh'}->prepare("SHOW TABLES LIKE '".$table."'");
        $sth->execute;
    };
    if($MainCfg->get('dbtype') eq 'Pg') {
        $sth=$self->{'dbh'}->prepare("SELECT * FROM pg_catalog.pg_class c LEFT JOIN pg_catalog.pg_user u ON u.usesysid = c.relowner LEFT JOIN pg_catalog.pg_namespace n ON n.oid = c.relnamespace WHERE c.relkind IN ('r','') AND n.nspname NOT IN ('pg_catalog', 'pg_toast') AND pg_catalog.pg_table_is_visible(c.oid) AND c.relname='".$table."'");
        $sth->execute;
    };
    if($sth->rows>0) {
        return(1);
    };
    return(0);
};

my %Datatype_mysql=(
              "varchar" =>  "VARCHAR(255)",
              "bigint"  =>  "BIGINT(20)",
              "int"     =>  "INT(11)",
              "bool"    =>  "INT(1)",
              "text"    =>  "TEXT",
              "enum"    =>  "ENUM",
              "date"    =>  "DATE",
              "decimal" =>  "DECIMAL(10,2)",
              "decimal4" =>  "DECIMAL(10,4)",
);

my %Datatype_Pg=(
              "varchar" =>  "VARCHAR(255) default ''",
              "bigint"  =>  "bigint default '0'",
              "int"     =>  "smallint default '0'",
              "bool"    =>  "bit default '0'",
              "text"    =>  "TEXT default ''",
              "enum"    =>  "CHAR(1) default ''",
              "date"    =>  "DATE",
              "decimal" =>  "FLOAT8 default '0'",
              "decimal4" =>  "FLOAT8 default '0'"
);

my %Datatype_Pg_update=(
              "varchar" =>  "VARCHAR(255)",
              "bigint"  =>  "bigint",
              "int"     =>  "smallint",
              "bool"    =>  "bit",
              "text"    =>  "TEXT",
              "enum"    =>  "CHAR(1)",
              "date"    =>  "DATE",
              "decimal" =>  "FLOAT8",
              "decimal4" =>  "FLOAT8"
);

my %Datatype_Pg_update_defaults=(
              "varchar" =>  "''",
              "bigint"  =>  "'0'",
              "int"     =>  "'0'",
              "bool"    =>  "'0'",
              "text"    =>  "''",
              "enum"    =>  "''",
              "date"    =>  undef,
              "decimal" =>  "'0'",
              "decimal4" =>  "'0'"
);


sub CreateTable {
    my($self, $table,$tdef)=@_;
    
    my $MainCfg=$MyMoney{'MainCfg'};
    my @tabledef=@{$tdef};
    my $sqlbatch="";
    my $keys="";
    my %Datatype;
    my @sequences=();
    my %create_key=();
    if($MainCfg->get('dbtype') eq 'mysql') {
        %Datatype=%Datatype_mysql;
    };
    if($MainCfg->get('dbtype') eq 'Pg') {
        %Datatype=%Datatype_Pg;
        $table=~tr/([A-Z])/([a-z])/;
    };
    foreach my $rdef (@tabledef) {
        my @rowdef=@{$rdef};
        #print("\t@rowdef\n");
        if($sqlbatch) {
            $sqlbatch.=", ";
        };
        if(($rowdef[2] eq "enum") && ($MainCfg->get('dbtype') eq 'mysql')) {
            my $enumvals="";
            for(my $i=0; $i<length($rowdef[3]);$i++) {
                if($enumvals) {
                    $enumvals.=",";
                };
                $enumvals.="'".substr($rowdef[3],$i,1)."'";
            };
            $sqlbatch.=$rowdef[0]." ".$Datatype{$rowdef[2]}."(".$enumvals.") NOT NULL";
        } else {
            if($MainCfg->get('dbtype') eq 'mysql') {
                $sqlbatch.=$rowdef[0]." ".$Datatype{$rowdef[2]}." NOT NULL";
                if($rowdef[3]) {
                    if($rowdef[3] eq "auto_increment") {
                        $sqlbatch.=" auto_increment";
                    } else {
                        if($rowdef[3] ne '') {
                            print("Error: neznamy atribut '".$rowdef[3]."'!\n");
                            exit;
                        };
                    };
                };
            };
            if($MainCfg->get('dbtype') eq 'Pg') {
                if($rowdef[2] eq 'enum') {
                    $rowdef[3]="";
                };
                if($rowdef[3]) {
                    if($rowdef[3] eq "auto_increment") {
                        $sqlbatch.=$rowdef[0]." SERIAL";
                    } else {
                        if(($rowdef[3] ne '') && ($rowdef[2] ne 'enum')) {
                            print("Error: neznamy atribut '".$rowdef[3]."'!\n");
                            exit;
                        };
                    };
                } else {
                    $sqlbatch.=$rowdef[0]." ".$Datatype{$rowdef[2]};
                };
            };
                
        };
        if($rowdef[4]) {
            if($MainCfg->get('dbtype') eq 'mysql') {
                if($keys) {
                    $keys.=", ";
                };
                if($rowdef[4] eq "primary") {
                    $keys.=" PRIMARY KEY(".$rowdef[0].")";
                } else {
                    if($rowdef[4] eq "key") {
                            $keys.=" KEY k_".$rowdef[0]."(".$rowdef[0].")";
                    } else {
                        if($rowdef[4] eq "unique") {
                            $keys.=" UNIQUE KEY k_".$rowdef[0]."(".$rowdef[0].")";
                        } else {
                            if($rowdef[4] ne '') {
                                print("Error: neznamy klic '".$rowdef[4]."'!\n");
                                exit;
                            };
                        };
                    };
                };
            };
            if($MainCfg->get('dbtype') eq 'Pg') {
                if($rowdef[4] eq "key") {
                    $create_key{$rowdef[0]}='N';
                } else {
                    #if(($rowdef[4] eq "unique") || ($rowdef[4] eq "primary")) {
                    if(($rowdef[4] eq "unique") || ($rowdef[4] eq "primary")) {
                        $create_key{$rowdef[0]}='U';
                    } else {
                        #if($rowdef[4] ne '') {
                        if($rowdef[4] ne ''){
                            print("Error: neznamy klic '".$rowdef[4]."'!\n");
                            exit;
                        };
                    };
                };
            };
            
        };
        
    };
    
    if(!$self->{'connected'}) {
        if(!$self->connect) {
            return(-1);
        };
    };
    my $local_transaction=0;
    if(!$self->{'transaction'}) {
        $self->begin_work;
        #$self->lock_tables($table);
        $local_transaction=1;
    };

    if($MainCfg->get('dbtype') eq 'mysql') {
        if($keys) {
            $sqlbatch.=", ".$keys;
        };
        my $sth=$self->{'dbh'}->do("CREATE TABLE ".$table." (".$sqlbatch.") TYPE=BerkeleyDB");
    };
    if($MainCfg->get('dbtype') eq 'Pg') {
        #foreach $_ (@sequences) {
        #    my $sth=$dbh->do("CREATE SEQUENCE \"".$table."_".$_."_seq\" start 1 increment 1 maxvalue 9223372036854775807 minvalue 1 cache 1");
        #};
        #print("CREATE TABLE ".$table." (".$sqlbatch.");\n");
        my $sth=$self->{'dbh'}->do("CREATE TABLE ".$table." (".$sqlbatch.")");
        foreach $_ (keys(%create_key)) {
            if($create_key{$_} eq "U") {
                my $sth=$self->{'dbh'}->do("CREATE UNIQUE INDEX ku_".$table."_".$_." ON ".$table." USING btree (".$_.")");
            } else {
                my $sth=$self->{'dbh'}->do("CREATE INDEX k_".$table."_".$_." ON ".$table." USING btree (".$_.")");
            };
        };
    };

    if($self->{'dbh'}->err) {
        if($local_transaction) {
            $self->rollback;
            #$self->unlock_tables($table);
        };
        return(0);
    };
    if($local_transaction) {
        $self->commit;
        #$self->unlock_tables($table);
    };
    return(1);
};

sub AddChangeColumn {
    my($self, $table,@rowdef)=@_;
    
    my $MainCfg=$MyMoney{'MainCfg'};
    my $sqlbatch="";
    my $keys="";
    my %Datatype;
    my @sequences=();
    my %create_key=();
    if($MainCfg->get('dbtype') eq 'mysql') {
        %Datatype=%Datatype_mysql;
    };
    if($MainCfg->get('dbtype') eq 'Pg') {
        %Datatype=%Datatype_Pg_update;
        $table=~tr/([A-Z])/([a-z])/;
    };

    #print("\t@rowdef\n");

    my $AlterAction="ADD";
    my $OldColName="";
    if($rowdef[0] eq 'change') {
       $AlterAction="CHANGE";
       $OldColName=$rowdef[1];
       shift(@rowdef);
    };
    shift(@rowdef);
    my $NewColName=$rowdef[0];
    
    if(($rowdef[2] eq "enum") && ($MainCfg->get('dbtype') eq 'mysql')) {
        my $enumvals="";
        for(my $i=0; $i<length($rowdef[3]);$i++) {
            if($enumvals) {
                $enumvals.=",";
            };
            $enumvals.="'".substr($rowdef[3],$i,1)."'";
        };
        $sqlbatch.=$rowdef[0]." ".$Datatype{$rowdef[2]}."(".$enumvals.") NOT NULL";
    } else {
        if($MainCfg->get('dbtype') eq 'mysql') {
            $sqlbatch.=$rowdef[0]." ".$Datatype{$rowdef[2]}." NOT NULL";
            if($rowdef[3]) {
                if($rowdef[3] eq "auto_increment") {
                    $sqlbatch.=" auto_increment";
                } else {
                    if($rowdef[3] ne '') {
                        print("Error: neznamy atribut '".$rowdef[3]."'!\n");
                        exit;
                    };
                };
            };
        };
        if($MainCfg->get('dbtype') eq 'Pg') {
            if($rowdef[2] eq 'enum') {
                $rowdef[3]="";
            };
            if($rowdef[3]) {
                if($rowdef[3] eq "auto_increment") {
                    $sqlbatch.=$rowdef[0]." SERIAL";
                } else {
                    if(($rowdef[3] ne '') && ($rowdef[2] ne 'enum')) {
                        print("Error: neznamy atribut '".$rowdef[3]."'!\n");
                        exit;
                    };
                };
            } else {
                $sqlbatch.=$rowdef[0]." ".$Datatype{$rowdef[2]};
            };
        };
            
    };
    my $Default="";
    my $NewDataType="";
    if($MainCfg->get('dbtype') eq 'Pg') {
        $Default=$Datatype_Pg_update_defaults{$rowdef[2]};
        $NewDataType=$Datatype_Pg_update{$rowdef[2]};
    };
    if($rowdef[4]) {
        if($rowdef[4] eq "key") {
            $create_key{$rowdef[0]}='N';
        } else {
            #if(($rowdef[4] eq "unique") || ($rowdef[4] eq "primary")) {
            if(($rowdef[4] eq "unique") || ($rowdef[4] eq "primary")) {
                $create_key{$rowdef[0]}='U';
            } else {
                #if($rowdef[4] ne '') {
                if($rowdef[4] ne ''){
                    print("Error: neznamy klic '".$rowdef[4]."'!\n");
                    exit;
                };
            };
        };
    };

    #if($rowdef[4]) {
    #    if($MainCfg->get('dbtype') eq 'mysql') {
    #        if($keys) {
    #            $keys.=", ";
    #        };
    #        if($rowdef[4] eq "primary") {
    #            $keys.=" PRIMARY KEY(".$rowdef[0].")";
    #        } else {
    #            if($rowdef[4] eq "key") {
    #                    $keys.=" KEY k_".$rowdef[0]."(".$rowdef[0].")";
    #            } else {
    #                if($rowdef[4] eq "unique") {
    #                    $keys.=" UNIQUE KEY k_".$rowdef[0]."(".$rowdef[0].")";
    #                } else {
    #                    if($rowdef[4] ne '') {
    #                        print("Error: neznamy klic '".$rowdef[4]."'!\n");
    #                        exit;
    #                    };
    #                };
    #            };
    #        };
    #    };
    #    if($MainCfg->get('dbtype') eq 'Pg') {
    #        if($rowdef[4] eq "key") {
    #            $create_key{$rowdef[0]}='N';
    #        } else {
    #            #if(($rowdef[4] eq "unique") || ($rowdef[4] eq "primary")) {
    #            if(($rowdef[4] eq "unique") || ($rowdef[4] eq "primary")) {
    #                $create_key{$rowdef[0]}='U';
    #            } else {
    #                #if($rowdef[4] ne '') {
    #                if($rowdef[4] ne ''){
    #                    print("Error: neznamy klic '".$rowdef[4]."'!\n");
    #                    exit;
    #                };
    #            };
    #        };
    #    };
    #    
    #};
    
    
    if(!$self->{'connected'}) {
        if(!$self->connect) {
            return(-1);
        };
    };
    my $local_transaction=0;
    if(!$self->{'transaction'}) {
        $self->begin_work;
        #$self->lock_tables($table);
        $local_transaction=1;
    };

    if($MainCfg->get('dbtype') eq 'mysql') {
        #print("ALTER TABLE ".$table." ".$AlterAction." COLUMN ".$OldColName." ".$sqlbatch.";\n");
        my $sth=$self->{'dbh'}->do("ALTER TABLE ".$table." ".$AlterAction." COLUMN ".$OldColName." ".$sqlbatch);

        foreach $_ (keys(%create_key)) {
            if($create_key{$_} eq "U") {
                #print("ALTER TABLE ".$table." ADD UNIQUE k_".$_."(".$_.")\n");
                my $sth=$self->{'dbh'}->do("ALTER TABLE ".$table." ADD UNIQUE k_".$_."(".$_.")");
            } else {
                #print("ALTER TABLE ".$table." ADD KEY k_".$_."(".$_.");\n");
                my $sth=$self->{'dbh'}->do("ALTER TABLE ".$table." ADD KEY k_".$_."(".$_.")");
            };
        };

    };
    if($MainCfg->get('dbtype') eq 'Pg') {
        #print("ALTER TABLE ".$table." ADD COLUMN ".$sqlbatch.";\n");
        my $sth=$self->{'dbh'}->do("ALTER TABLE ".$table." ADD COLUMN ".$sqlbatch);
        if($Default) {
            #print("ALTER TABLE ".$table." ALTER COLUMN ".$NewColName." SET DEFAULT ".$Default.";\n");
            my $sth=$self->{'dbh'}->do("ALTER TABLE ".$table." ALTER COLUMN ".$NewColName." SET DEFAULT ".$Default);
        };
        if($AlterAction eq "CHANGE") {
            #print("UPDATE ".$table." SET ".$NewColName." = CAST(".$OldColName." AS ".$NewDataType.");\n");
            my $sth=$self->{'dbh'}->do("UPDATE ".$table." SET ".$NewColName." = CAST(".$OldColName." AS ".$NewDataType.")");
            #print("ALTER TABLE ".$table." DROP COLUMN ".$OldColName.";\n");
            $sth=$self->{'dbh'}->do("ALTER TABLE ".$table." DROP COLUMN ".$OldColName);
        };
        foreach $_ (keys(%create_key)) {
            if($create_key{$_} eq "U") {
                #print("CREATE UNIQUE INDEX ku_".$table."_".$_." ON ".$table." USING btree (".$_.");\n");
                my $sth=$self->{'dbh'}->do("CREATE UNIQUE INDEX ku_".$table."_".$_." ON ".$table." USING btree (".$_.")");
            } else {
                #print("CREATE INDEX k_".$table."_".$_." ON ".$table." USING btree (".$_.")\n");
                my $sth=$self->{'dbh'}->do("CREATE INDEX k_".$table."_".$_." ON ".$table." USING btree (".$_.")");
            };
        };
        #print("VACUUM FULL ".$table.";\n");
        #$sth=$self->{'dbh'}->do("VACUUM FULL ".$table);
    };

    if($self->{'dbh'}->err) {
        if($local_transaction) {
            $self->rollback;
            #$self->unlock_tables($table);
        };
        return(0);
    };
    if($local_transaction) {
        $self->commit;
        #$self->unlock_tables($table);
    };
    return(1);
};

sub InsertFromFile {
    my($self, $table,$file,@tbldef)=@_;

    my $MainCfg=$MyMoney{'MainCfg'};
    if($MainCfg->get('dbtype') eq 'Pg') {
        $table=~tr/([A-Z])/([a-z])/;
    };
    my $cols=0;
    my $tblstruct="";
    foreach $_ (@tbldef) {
        my @rowdef=@{$_};
        if(!$rowdef[3]) {
            $rowdef[3]="";
        };
        if($rowdef[3] ne "auto_increment") {
            if($tblstruct) {
                $tblstruct.=", ";
            };
            $tblstruct.=$rowdef[0];
            $cols++;
        };
    };

    if(!$self->{'connected'}) {
        if(!$self->connect) {
            return(-1);
        };
    };
    my $local_transaction=0;
    if(!$self->{'transaction'}) {
        $self->begin_work;
        $self->lock_tables($table);
        $local_transaction=1;
    };

    open(DATAFILE,$file) or die("Error: nelze otevrit soubor '".$file."'!\n");
    while(<DATAFILE>) {
        chomp;
        #$_=Encode::encode('iso-8859-2', $_);
        my @row=split(';',$_);
        my $i=0;
        my $sqlbatch="";
        foreach $_ (@row) {
            if($sqlbatch) {
                $sqlbatch.=", ";
            };
            $sqlbatch.="'".$_."'";
            $i++;
        };
        if($i<$cols) {
            for($i=$i;$i<$cols;$i++) {
                $sqlbatch.=", ''";
            };
        };
        #print("INSERT INTO ".$table." (".$tblstruct.") VALUES(".$sqlbatch.");\n");
        my $sth=$self->{'dbh'}->do("INSERT INTO ".$table." (".$tblstruct.") VALUES(".$sqlbatch.")");
        #print("\nINSERT INTO ".$table." VALUES(".$sqlbatch.");\n");
        if($self->{'dbh'}->err) {
            close(DATAFILE);
            if($local_transaction) {
                $self->rollback;
                $self->unlock_tables($table);
            };
           return(0);
        };
    };
    close(DATAFILE);
    if($local_transaction) {
        $self->commit;
        $self->unlock_tables($table);
    };
    return(1);
};

sub odstranit_rok {
    my($self, $rok_id)=@_;
    my $MainCfg=$MyMoney{'MainCfg'};
    my $tablename="g_Rok";


    if(!$self->{'connected'}) {
        if(!$self->connect) {
            return(-1);
        };
    };
    my $local_transaction=0;
    if(!$self->{'transaction'}) {
        $self->begin_work;
        $self->lock_tables($tablename);
        $local_transaction=1;
    };

    my $sth;
    my $sth_drop;
    
    if($MainCfg->get('dbtype') eq 'mysql') {
        $sth=$self->{'dbh'}->prepare("SHOW TABLES LIKE 'r".$rok_id."_%'");
        $sth->execute;
    };
    if($MainCfg->get('dbtype') eq 'Pg') {
        $sth=$self->{'dbh'}->prepare("SELECT c.relname FROM pg_catalog.pg_class c LEFT JOIN pg_catalog.pg_user u ON u.usesysid = c.relowner LEFT JOIN pg_catalog.pg_namespace n ON n.oid = c.relnamespace WHERE c.relkind IN ('r','') AND n.nspname NOT IN ('pg_catalog', 'pg_toast') AND pg_catalog.pg_table_is_visible(c.oid) AND c.relname ~ 'r".$rok_id."_'");
        $sth->execute;
    };
    for(my $i=0;$i<$sth->rows;$i++) {
        my @row=$sth->fetchrow;
        my $sth_drop=$self->{'dbh'}->prepare("DROP TABLE ".$row[0]);
        $sth_drop->execute;
        if($self->{'dbh'}->err) {
            if($local_transaction) {
                $self->rollback;
                $self->unlock_tables($tablename);
            };
            return(0);
        };
    };
    if($MainCfg->get('dbtype') eq 'Pg') {
        $tablename=~tr/([A-Z])/([a-z])/;
    };
    $sth_drop=$self->{'dbh'}->prepare("DELETE FROM ".$tablename." WHERE id='".$rok_id."'");
    $sth_drop->execute;
    if($self->{'dbh'}->err) {
        if($local_transaction) {
            $self->rollback;
            $self->unlock_tables($tablename);
        };
        return(0);
    };
    if($local_transaction) {
        $self->commit;
        $self->unlock_tables($tablename);
    };
    return(1);
};

sub odstranit_agendu {
    my($self, $agenda_id)=@_;
    my $MainCfg=$MyMoney{'MainCfg'};
    my $tablename_rok="g_Rok";
    my $tablename_agenda="g_Agenda";
    my $sth_drop;

    if($MainCfg->get('dbtype') eq 'Pg') {
        $tablename_rok=~tr/([A-Z])/([a-z])/;
        $tablename_agenda=~tr/([A-Z])/([a-z])/;
    };

    if(!$self->{'connected'}) {
        if(!$self->connect) {
            return(-1);
        };
    };
    my $local_transaction=0;
    if(!$self->{'transaction'}) {
        $self->begin_work;
        $self->lock_tables($tablename_rok,$tablename_agenda);
        $local_transaction=1;
    };

    my $sth;
    $sth=$self->{'dbh'}->prepare("SELECT id FROM ".$tablename_rok." WHERE agenda_id='".$agenda_id."'");
    $sth->execute;
    for(my $i=0;$i<$sth->rows;$i++) {
        my @row=$sth->fetchrow;
        if(!$self->odstranit_rok($row[0])) {
            if($local_transaction) {
                $self->rollback;
                $self->unlock_tables($tablename_rok,$tablename_agenda);
            };
            return(0);
        };
    };
    
    if($MainCfg->get('dbtype') eq 'mysql') {
        $sth=$self->{'dbh'}->prepare("SHOW TABLES LIKE 'a".$agenda_id."_%'");
        $sth->execute;
    };
    if($MainCfg->get('dbtype') eq 'Pg') {
        $sth=$self->{'dbh'}->prepare("SELECT c.relname FROM pg_catalog.pg_class c LEFT JOIN pg_catalog.pg_user u ON u.usesysid = c.relowner LEFT JOIN pg_catalog.pg_namespace n ON n.oid = c.relnamespace WHERE c.relkind IN ('r','') AND n.nspname NOT IN ('pg_catalog', 'pg_toast') AND pg_catalog.pg_table_is_visible(c.oid) AND c.relname ~ 'a".$agenda_id."_'");
        $sth->execute;
    };
    for(my $i=0;$i<$sth->rows;$i++) {
        my @row=$sth->fetchrow;
        my $sth_drop=$self->{'dbh'}->prepare("DROP TABLE ".$row[0]);
        $sth_drop->execute;
        if($self->{'dbh'}->err) {
            if($local_transaction) {
                $self->rollback;
                $self->unlock_tables($tablename_rok,$tablename_agenda);
            };
            return(0);
        };
    };
    $sth_drop=$self->{'dbh'}->prepare("DELETE FROM ".$tablename_agenda." WHERE id='".$agenda_id."'");
    $sth_drop->execute;
    if($self->{'dbh'}->err) {
        if($local_transaction) {
            $self->rollback;
            $self->unlock_tables($tablename_rok,$tablename_agenda);
        };
        return(0);
    };
    if($local_transaction) {
        $self->commit;
        $self->unlock_tables($tablename_rok,$tablename_agenda);
    };
    return(1);
};

1;
