#!/usr/bin/perl -w

package Table;

use strict;

#use Encode;
use utf8;
use open ':utf8', ':std';

use POSIX qw(strftime);

use lib "./";
use lib::TableDef;
use lib::Databaze;
use lib::ProgressDialog;

my %MyMoney;

sub new {
  my $class = shift;
  my $self = {};
  bless $self, $class;
  $self->{'tablename'}=shift;
  %MyMoney=@_;
  if($self->{'tablename'}) {
      $self->_initialize;
  };
  $self->{'DB'}=0;
  return $self;
};

sub _initialize {
  my($self)=shift;
#  my @tabledef=@{$Table{$self->{'tablename'}}};
  my @tabledef=@{TableDef::get($self->{'tablename'})};
  foreach $_ (@tabledef) {
	my @row=@{$_};
	$self->{'Data'}{$row[0]}=$row[1];
  };
  $self->{'last_primary_value'}="";
};

sub set_db {
    my ($self,$DB)=@_;
    $self->{'DB'}=$DB;
};

sub get {
    my($self)=shift;
    my($key)=@_;
#    if(!$self->{'Data'}{$key}) {
#	my @tabledef=@{$Table{$self->{'tablename'}}};
#	foreach $_ (@tabledef) {
#		my @row=@{$_};
#		if($row[0] eq $key) {
#			if(($row[2] eq 'bigint') || ($row[2] eq 'int') || ($row[2] eq 'decimal')) {
#				return(0);
#			} else {
#				return('');
#			};
#		};
#	};
#    };
    return($self->{'Data'}{$key});
};

sub get_data {
    my($self)=shift;
    my @data=();
    foreach $_ (keys(%{$self->{'Data'}})) {
	push(@data,$_);
	push(@data,$self->get($_));
    };
    return(@data);
};

sub set_data {
    my($self,%data)=@_;
    $self->set(%data);
};

sub set {
    my($self)=shift;
    my(%values)=@_;

    foreach $_ (keys(%values)) {
	if(($self->get_auto eq '') && ($self->get_primary eq $_)) {
		if($self->{'last_primary_value'} eq '') {
			$self->{'last_primary_value'}=$self->get($_);
		};
	};
        $self->{'Data'}{$_}=$values{$_};
    };
};

sub unset {
	my($self,$key)=@_;
	delete($self->{'Data'}{$key});
};

sub get_auto {
	my($self)=shift;
	my @tabledef=@{TableDef::get($self->{'tablename'})};
	foreach $_ (@tabledef) {
	    my @row=@{$_};
	    if($row[3]) {
		if($row[3] eq "auto_increment") {
			return($row[0]);
		    };
	    };
	};
	return('');
};

sub get_primary {
	my($self)=shift;
	my @tabledef=@{TableDef::get($self->{'tablename'})};
	foreach $_ (@tabledef) {
	    my @row=@{$_};
	    if($row[4]) {
		if($row[4] eq "primary") {
			return($row[0]);
		    };
	    };
	};
	return('');
};

sub get_columns_wihout_auto {
	my($self)=shift;
	my @tabledef=@{TableDef::get($self->{'tablename'})};
	my @keys=();
	foreach $_ (@tabledef) {
	    my @row=@{$_};
	    if($row[3]) {
		if($row[3] ne "auto_increment") {
			push(@keys,$row[0]);
		};
	    } else {
		push(@keys,$row[0]);
	    };
	};
	return(@keys);
};

sub realtablename {
	my($self)=shift;
	
	my $tablename=$self->{'tablename'};
	if($tablename=~/^a_/) {
		my $prefix="a".$MyMoney{'MainCfg'}->get('agenda')."_";
		$tablename=~s/^a_/${prefix}/;
	};
	if($tablename=~/^r_/) {
		my $prefix="r".$MyMoney{'MainCfg'}->get('rok')."_";
		$tablename=~s/^r_/${prefix}/;
	};
	return($tablename);
};

sub get_column_info {
	my($self)=shift;
	my($column)=@_;
	my @tabledef=@{TableDef::get($self->{'tablename'})};
	my @keys=();
	foreach $_ (@tabledef) {
	    my @row=@{$_};
	    if($row[0] eq $column) {
		return(@row);
	    };
	};
	return(0);
};

sub save {
	my($self,$ignore_err)=@_;
	my %d=%{$self->{'Data'}};
	my %data=();
	my $Databaze;
	if(!$self->{'DB'}) {
		$Databaze=new Databaze(%MyMoney);
	} else {
		$Databaze=$self->{'DB'};
	};
	foreach $_ ($self->get_columns_wihout_auto) {
		my @col=$self->get_column_info($_);
		if($col[2]=~/^decimal/) {
			my $decval=$d{$_};
			$decval=~s/,/\./;
			$data{$_}=$decval;
		} else {
			$data{$_}=$d{$_};
		};
	};
	my $i=$Databaze->Insert($self->realtablename,$self->get_auto,%data);
	
	my $err=0;
	if($i) {
		if($i<0) {
			$err=1;
		};
	};
	if($err) {
	    if($ignore_err) {
		return(0);
	    };
	    print("Table->save: nelze ulozit data do tabulky '".$self->realtablename."'!\n");
	    exit;
	};
	$self->{'last_primary_value'}="";
	if($self->get_auto) {
		$self->set($self->get_auto,$i);
	};
	return(1);
};

sub update {
	my($self)=shift;
	my %data=%{$self->{'Data'}};
	my $Databaze;
	if(!$self->{'DB'}) {
		$Databaze=new Databaze(%MyMoney);
	} else {
		$Databaze=$self->{'DB'};
	};

	my $where="";
	if((my $auto=$self->get_auto)) {
		$where=$auto."=".$self->get($auto);
	} else {
		if((my $primary=$self->get_primary)) {
			if($self->{'last_primary_value'}) {
				$where=$primary."='".$self->{'last_primary_value'}."'";
			} else {
				$where=$primary."='".$self->get($primary)."'";
			};
		};
	};
	foreach $_ (keys(%data)) {
		my @col=$self->get_column_info($_);
		#print("i=$_\n");
		if($col[2]=~/^decimal/) {
			$data{$_}=~s/,/\./;
		};
	};
	if($where) {
		if($Databaze->Update($self->realtablename,$where,%data)>0) {
			$self->{'last_primary_value'}="";
		} else {
			my $Dialog=new Dialog;
			$Dialog->show('text'=>"Nezdarila se aktualizace zaznamu.\nZrejme se snazite ulozit duplicitni klic, nebo byl tento zaznam jiz jinym uzivatelem vymazan.",'type'=>'error');
			return(0);
		};
	} else {
	    print("Table->update: pokus o aktualizaci vsech dat v tabulce '".$self->realtablename."'!\n");
	    exit;
	};
	return(1);
};

sub load {
	my($self,$where,$ignore_error)=@_;

	my $Databaze;
	if(!$self->{'DB'}) {
		$Databaze=new Databaze(%MyMoney);
	} else {
		$Databaze=$self->{'DB'};
	};

	my %data=%{$self->{'Data'}};
	if($Databaze->Read($self->realtablename,$where,keys(%data))) {
		my %data=$Databaze->fetchhash;
		foreach $_ (keys(%data)) {
			$self->{'Data'}{$_}=$data{$_};
		};
		if($self->get_auto eq '' && (my $primary=$self->get_primary)) {
			$self->{'last_primary_value'}=$self->get($primary);
		};
	} else {
		print("Table->load: nelze nacist data z tabulky '".$self->realtablename."'; where=\"".$where."\" !\n");
		if(!$ignore_error) {
			exit;
		} else {
			return(0);
		};
	};
	return(1);
};

my $realregexp;
my $regexp;

sub check {
    my($self)=shift;
    ($realregexp)=@_;
    my @check_tables=();
    $regexp=$realregexp;
    if($regexp =~ /[0-9]/) {
	    $regexp=~s/[0-9]//g;
    };
    foreach $_ (TableDef::get_table_names) {
        if(/${regexp}/) {
            push(@check_tables,$_);
        };
    };
    $self->{'check_tables'}=[@check_tables];
    my $i=@check_tables;
    if($i) {
	my $label="Kontrola globálních tabulek ...";
	if($regexp =~ /^\^a/) {
		$label="Kontrola tabulek agendy ...";
	};
	if($regexp =~ /^\^r/) {
		$label="Kontrola tabulek účetního roku ...";
	};
        my $Progress=new ProgressDialog;
        $Progress->run("label"=>1,"steps"=>$i,
        	       "callback"=>$self,
		       "timeout"=>200,
		       "callback_sub"=>"check_tables_cb",
		       "toplabel"=>$label);
    };
    if($regexp =~ /^\^a/) {
	$MyMoney{'AgendaData'}->set('db_version',TableDef::getDBversion);
	$MyMoney{'AgendaData'}->update;
    };
    if($regexp =~ /^\^r/) {
	$MyMoney{'RokData'}->set('db_version',TableDef::getDBversion);
	$MyMoney{'RokData'}->update;
    };
    return(1);
}

sub check_tables_cb {
	my($self,$Progress,$val)=@_;
	my $table=$self->{'check_tables'}[$val];
	my $realtable=$table;
	my $Databaze;
	if(!$self->{'DB'}) {
		$Databaze=new Databaze(%MyMoney);
	} else {
		$Databaze=$self->{'DB'};
	};

	if($regexp ne $realregexp) {
		$realtable=~s/..//;
		$_=$realregexp;
		s/\^//;
		$realtable=$_.$realtable;
	};
        my $i=$Databaze->Check($realtable);
        if($i<0) {
            print("Error: nezdarila se kontrola tabulky '".$realtable."'!\n");
            exit;
        };
        if($i==0) {
            print("Vytvarim tabulku '".$realtable."' ...\n");
            if(!$Databaze->CreateTable($realtable,TableDef::get($table))) {
                print("Error: neporadilo se vytvorit tabulku!\n");
                exit;
            };
	    my $tabledir="";
	    if($table=~/^g/) {
		$tabledir="./data/global/";
	    };
	    if($table=~/^a/) {
		$tabledir="./data/agenda/";
	    };
	    if($table=~/^r/) {
		$tabledir="./data/rok/".$MyMoney{'RokData'}->get('ucetni_rok')."/";
	    };
	    if($tabledir) {
		if( -e $tabledir.$table) {
		    print("\t- nacitam zakladni data\n");
		    my @tbldef=@{TableDef::get($table)};
		    if(!$Databaze->InsertFromFile($realtable,$tabledir.$table,@tbldef)) {
			print("Error: nepodarilo se nacist data ze souboru '".$tabledir.$table."'!\n");
			exit;
		    };
		};
	    };
        } else {
		#
		# Tabulka existuje, ale jeste zkontrolujeme verzi
		#
		if($table=~/^g/) {
			if($table=~/^g_Agenda/) {
				$Databaze->Read('g_Agenda',"","*");
				my %row=$Databaze->fetchhash;
				if(!defined($row{'db_version'})) {
					my $tbversion="0.0.0";
					$self->UpdateVersion($table,$realtable,$tbversion);
				};
			};
			if($table=~/^g_Rok/) {
				$Databaze->Read('g_Rok',"","*");
				my %row=$Databaze->fetchhash;
				if(!defined($row{'db_version'})) {
					my $tbversion="0.0.0";
					$self->UpdateVersion($table,$realtable,$tbversion);
				};
			};
		};
		if($table=~/^a/) {
			if($MyMoney{'AgendaData'}->get('db_version') ne TableDef::getDBversion) {
				$self->UpdateVersion($table,$realtable,$MyMoney{'AgendaData'}->get('db_version'));
			};
		};
		if($table=~/^r/) {
			if($MyMoney{'RokData'}->get('db_version') ne TableDef::getDBversion) {
				$self->UpdateVersion($table,$realtable,$MyMoney{'RokData'}->get('db_version'));
			};
		};
	};
	return(1);
};

sub UpdateVersion {
	my($self,$table,$realtable,$tbversion)=@_;

	my $Databaze;
	if(!$self->{'DB'}) {
		$Databaze=new Databaze(%MyMoney);
	} else {
		$Databaze=$self->{'DB'};
	};

	my $in_update=0;
	foreach my $version (TableDef::getDBversions) {
		if($in_update) {
			my %DBupdates=%{TableDef::getDBupdates($version)};
			if(defined($DBupdates{$table})) {
				print("Aktualizuji tabulku $realtable na DB verzi $version\n");
				foreach my $row (@{$DBupdates{$table}}) {
					my @alterdata=@{$row};
					if(($alterdata[0] eq "add") || ($alterdata[0] eq "change")) {
						$Databaze->AddChangeColumn($realtable,@alterdata);
					} elsif($alterdata[0] eq "set") {
						shift(@alterdata);
						$Databaze->Update($realtable,@alterdata);
					};

				};
			};
		};
		if($version eq $tbversion) {
			$in_update=1;
		};
	};
};

#sub get_tablenames {
#	return(keys(%Table));
#};

my $debug=0;

sub ReadFromUI {
	my($self,$f)=@_;
	my %form=%{$f};
	my @tbstruct=@{TableDef::get($self->{'tablename'})};
	foreach $_ (@tbstruct) {
		my @rstruct=@{$_};
		if($rstruct[2] ne "enum") {
			if($form{$rstruct[0]}) {
				if(ref($form{$rstruct[0]}) eq "Gtk2::Entry") {
					$self->set($rstruct[0],$form{$rstruct[0]}->get_text);
				} else {
					if(ref($form{$rstruct[0]}) eq "Gtk2::SpinButton") {
						my $decval=$form{$rstruct[0]}->get_value;
						$decval=~s/,/\./;
						$self->set($rstruct[0],$decval);
					} else {
						if(ref($form{$rstruct[0]}) eq "Gtk2::CheckButton") {
							if($form{$rstruct[0]}->get_active) {
								$self->set($rstruct[0],1);
							} else {
								$self->set($rstruct[0],0);
							}
						} else {
							if(ref($form{$rstruct[0]}) eq "Gtk2::TextView") {
								$self->set($rstruct[0],$form{$rstruct[0]}->get_buffer->get_text($form{$rstruct[0]}->get_buffer->get_start_iter,$form{$rstruct[0]}->get_buffer->get_end_iter,1));
							} else {
								if(ref($form{$rstruct[0]}) eq "Gtk2::OptionMenu") {
									$self->set($rstruct[0],$form{$rstruct[0]}->get_history);
								} else {
									if(ref($form{$rstruct[0]}) eq "Gnome2::DateEdit") {
										my $datum=strftime("%Y-%m-%d",localtime($form{$rstruct[0]}->get_time));
										if($datum eq "1970-01-01") {
											$datum="0001-01-01";
										};
										$self->set($rstruct[0],$datum);
									} else {
										if(ref($form{$rstruct[0]}) eq "Gtk2::Label") {
											if($rstruct[2]=~/^decimal/) {
												my $decval=$form{$rstruct[0]}->get_label;
												$decval=~s/,/\./;
												$self->set($rstruct[0],$decval);
											} else {
												$self->set($rstruct[0],$form{$rstruct[0]}->get_label);
											};
										} else {
											print("Debug: neznamy typ '".ref($form{$rstruct[0]})."'!\n");
										};
									};
								};
							};
						};
					};
				};
			} else {
				if($debug) {
					print("Debug: polozka '".$rstruct[0]."' neni obsazena ve formulari!\n");
				};
			};
		} else {
			for(my $i=0; $i<length($rstruct[3]);$i++) {
				my $enumname=$rstruct[0]."_".substr($rstruct[3],$i,1);
				if($form{$enumname}) {
					if(ref($form{$enumname}) eq "Gtk2::RadioButton") {
						if($form{$enumname}->get_active) {
							$self->set($rstruct[0],substr($rstruct[3],$i,1));
						};
					} else {
						print("Debug: neznamy typ '".ref($form{$enumname})."' - byl ocekavan Gtk2::RadioButton!\n");
					};
				} else {
					if($debug) {
						print("Debug: polozka '".$enumname."' neni obsazena ve formulari!\n");
					};
				};
			};

		};
	};
};

sub WriteToUI {
	my($self,$f)=@_;
	my %form=%{$f};
	my @tbstruct=@{TableDef::get($self->{'tablename'})};
	foreach $_ (@tbstruct) {
		my @rstruct=@{$_};
		if($rstruct[2] ne "enum") {
			if($form{$rstruct[0]}) {
				if(ref($form{$rstruct[0]}) eq "Gtk2::Entry") {
					$form{$rstruct[0]}->set_text($self->get($rstruct[0]));
				} else {
					if(ref($form{$rstruct[0]}) eq "Gtk2::SpinButton") {
						$form{$rstruct[0]}->set_value($self->get($rstruct[0]));
					} else {
						if(ref($form{$rstruct[0]}) eq "Gtk2::CheckButton") {
							$form{$rstruct[0]}->set_active($self->get($rstruct[0]));
						} else {
							if(ref($form{$rstruct[0]}) eq "Gtk2::TextView") {
								$form{$rstruct[0]}->get_buffer->set_text($self->get($rstruct[0]));
							} else {
								if(ref($form{$rstruct[0]}) eq "Gtk2::OptionMenu") {
									$form{$rstruct[0]}->set_history($self->get($rstruct[0]));
								} else {
									if(ref($form{$rstruct[0]}) eq "Gnome2::DateEdit") {
										my @d=split('-',$self->get($rstruct[0]));
										my @datum=localtime;
										$datum[3]=$d[2];
										$datum[4]=$d[1]-1;
										$datum[5]=$d[0]-1900;
										$form{$rstruct[0]}->set_time(strftime('%s',@datum));
									} else {
										if(ref($form{$rstruct[0]}) eq "Gtk2::Label") {
											if($rstruct[2] eq "decimal") {
												$form{$rstruct[0]}->set_label(sprintf("%01.2f",$self->get($rstruct[0])));
											} else {
												if($rstruct[2] eq "decimal4") {
													$form{$rstruct[0]}->set_label(sprintf("%01.4f",$self->get($rstruct[0])));
												} else {
													$form{$rstruct[0]}->set_label($self->get($rstruct[0]));
												};
											};
										} else {
											print("Debug: neznamy typ '".ref($form{$rstruct[0]})."'!\n");
										};
									};
								};
							};
						};
					};
				};
			} else {
				if($debug) {
					print("Debug: polozka '".$rstruct[0]."' neni obsazena ve formulari!\n");
				};
			};
		} else {
			my $enumname=$rstruct[0]."_".$self->get($rstruct[0]);
			if($form{$enumname}) {
				if(ref($form{$enumname}) eq "Gtk2::RadioButton") {
					$form{$enumname}->set_active(1);
				} else {
					print("Debug: neznamy typ '".ref($form{$enumname})."' - byl ocekavan Gtk2::RadioButton!\n");
				};
			} else {
				if($debug) {
					print("Debug: polozka '".$enumname."' neni obsazena ve formulari!\n");
				};
			};
		};
	};
};

sub read_defaults {
	my($self,$rok)=@_;
	
	my $filename="./data/rok/".$rok."/".$self->{'tablename'}.".default";
	if(-e $filename) {
		open(FILE,$filename);
		while(<FILE>) {
			chomp;
			my ($key,$value)=split('=',$_);
			$self->set($key,$value);
		};
		close(FILE);
	};
};

sub read_seznamy_defaults {
	my($self,$rok)=@_;
	
	my $filename="./data/rok/".$rok."/".$self->{'tablename'}.".seznamy";
	if(-e $filename) {
		open(FILE,$filename);
		while(<FILE>) {
			chomp;
			my ($key,$value)=split('=',$_);
			$self->set($key,$value);
		};
		close(FILE);
	};
};

sub createdb {
	my($self)=shift;
	
	my $id=$self->{'Data'}{'id'};
	my @files=`ls -1 ./sql/a_*`;
	$self->{'sql_files'}=[@files];
	my $Progress=new ProgressDialog;
	my $i=@files;
	$Progress->run("steps"=>$i,
		       "callback"=>$self,
		       "timeout"=>300,
		       "callback_sub"=>"create_tables_from_sql_file",
		       "toplabel"=>"Vytvářím tabulky nové agendy ...");

};

sub create_tables_from_sql_file {
	my($self,$Progress,$val)=@_;

	my $file=$self->{'sql_files'}[$val];
	chomp($file);
	my $table=$file;
	$table=~s/^\.\/sql\///;
	$table=~s/\.sql$//;
	my $newtable=$table;
	my $id=$self->get('id');
	$newtable=~s/^a/a${id}/;
	my @i=@{$self->{'sql_files'}};
	my $percent=sprintf("%s - %i %%",$newtable,$val*(1/@i)*100);
	$Progress->set_progress_label($percent);
	#print("$file\t$table\t$newtable\n");
	open(TABLEFILE,$file);
	my $sqlbatch="";
	my $Databaze;
	if(!$self->{'DB'}) {
		$Databaze=new Databaze(%MyMoney);
	} else {
		$Databaze=$self->{'DB'};
	};

	while(<TABLEFILE>) {
		while(/ ${table} /) {
			s/ ${table} / ${newtable} /;
		};
		$sqlbatch.=$_;
		if(/;/) {
			$Databaze->Create($sqlbatch);
			$sqlbatch="";
		};
	};
	close(TABLEFILE);

	return(1);
};

1;
